procedure ComplexConvolution(NumPoints : integer;
                         var XReal     : TNvectorPtr;
                         var XImag     : TNvectorPtr;
                         var HReal     : TNvectorPtr;
                         var HImag     : TNvectorPtr;
                         var Error     : byte);

{-------------------------------------------------------------------}
{-                                                                 -}
{-     Turbo Pascal Numerical Methods Toolbox                      -}
{-     (C) Copyright 1986 Borland International.                   -}
{-                                                                 -}
{-   Input: NumPoints, XReal, XImag, HReal, HImag                  -}
{-   Output: XReal, XImag, Error                                   -}
{-                                                                 -}
{-   Purpose: This procedure performs a convolution of the         -}
{-            data XReal, XImag and the data HReal and HImag. The  -}
{-            vectors XReal, XImag, HReal and HImag are            -}
{-            transformed in place.                                -}
{-                                                                 -}
{- User Defined Types:                                             -}
{-         TNvector = array[0..TNArraySize] of real                -}
{-      TNvectorPtr = ^TNvector                                    -}
{-                                                                 -}
{- Global Variables:  NumPoints   : integer     Number of data     -}
{-                                              points in X        -}
{-                    XReal,XImag : TNvectorPtr Data points        -}
{-                    HReal,HImag : TNvectorPtr Data points        -}
{-                    Error       : byte        Indicates an error -}
{-                                                                 -}
{-             Errors:  0: No Errors                               -}
{-                      1: NumPoints < 2                           -}
{-                      2: NumPoints not a power of two            -}
{-                                                                 -}
{-   Version Date:  26 January 1987                                -}
{-                                                                 -}
{-------------------------------------------------------------------}

var
  SinTable, CosTable : TNvectorPtr;    { Tables of sine and cosine values  }
  Inverse : boolean;                   { Indicates inverse transform  }
  NumberOfBits : byte;                 { # of bits required to  }
                                       { represent NumPoints    }

procedure Multiply(NumPoints : integer;
               var XReal     : TNvectorPtr;
               var XImag     : TNvectorPtr;
               var HReal     : TNvectorPtr;
               var HImag     : TNvectorPtr);

{----------------------------------------------}
{- Input: NumPoints, XReal, XImag, HReal,     -}
{-        HImag                               -}
{- Output: XReal, XImag                       -}
{-                                            -}
{- This procedure multiplies each element in  -}
{- X by the corresponding element in H.  The  -}
{- product is returned in X.                  -}
{----------------------------------------------}

var
  Term : integer;
  Dummy : real;

begin
  for Term := 0 to NumPoints - 1 do
  begin
    Dummy := XReal^[Term] * HReal^[Term] - XImag^[Term] * HImag^[Term];
    XImag^[Term] := XReal^[Term] * HImag^[Term] +
                    XImag^[Term] * HReal^[Term];
    XReal^[Term] := Dummy;
  end;
end; { procedure Multiply }

begin { procedure ComplexConvolution }
  TestInput(NumPoints, NumberOfBits, Error);
  if Error = 0 then
  begin
    New(SinTable);
    New(CosTable);
    MakeSinCosTable(NumPoints, SinTable, CosTable);

    { Take the transform of XReal, XImag  }
    Inverse := false;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);

    { Take the transform of HReal, HImag  }
    FFT(NumberOfBits, NumPoints, Inverse, HReal, HImag, SinTable, CosTable);

    { Multiply the two transforms  }
    Multiply(NumPoints, XReal, XImag, HReal, HImag);

    { Take the inverse transform of the product  }
    Inverse := true;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);
    Dispose(SinTable);
    Dispose(CosTable);
  end;
end; { procedure ComplexConvolution }
