procedure ComplexCorrelation(NumPoints : integer;
                         var Auto      : boolean;
                         var XReal     : TNvectorPtr;
                         var XImag     : TNvectorPtr;
                         var HReal     : TNvectorPtr;
                         var HImag     : TNvectorPtr;
                         var Error     : byte);

{-------------------------------------------------------------------}
{-                                                                 -}
{-     Turbo Pascal Numerical Methods Toolbox                      -}
{-     (C) Copyright 1986 Borland International.                   -}
{-                                                                 -}
{-   Input: NumPoints, Auto, XReal, XImag, HReal, HImag            -}
{-   Output: XReal, XImag, Error                                   -}
{-                                                                 -}
{-   Purpose: This procedure performs a correlation (auto or       -}
{-            cross) of the complex data XReal, XImag and the      -}
{-            complex data HReal, HImag. The vectors XReal, XImag, -}
{-            HReal, and HImag are transformed in place.           -}
{-                                                                 -}
{- User Defined Types:                                             -}
{-         TNvector = array[0..TNArraySize] of real                -}
{-      TNvectorPtr = ^TNvector                                    -}
{-                                                                 -}
{- Global Variables:  NumPoints   : integer   Number of data       -}
{-                                            points in X          -}
{-                    Auto        : boolean   True => auto-        -}
{-                                                    correlation  -}
{-                                            False=> cross-       -}
{-                                                    correlation  -}
{-                    XReal,XImag : TNvectorPtr First sample       -}
{-                    HReal,HImag : TNvectorPtr Second sample      -}
{-                    Error       : byte        Indicates an error -}
{-                                                                 -}
{-             Errors:  0: No Errors                               -}
{-                      1: NumPoints < 2                           -}
{-                      2: NumPoints not a power of two            -}
{-                                                                 -}
{-   Version Date: 26 January 1987                                 -}
{-                                                                 -}
{-------------------------------------------------------------------}

var
  SinTable, CosTable : TNvectorPtr;  { Tables of sine and cosine values  }
  Inverse : boolean;                 { Indicates inverse transform  }
  NumberOfBits : byte;               { Number of bits necessary to     }
                                     { represent the number of points  }

procedure Multiply(NumPoints : integer;
               var XReal     : TNvectorPtr;
               var XImag     : TNvectorPtr;
               var HReal     : TNvectorPtr;
               var HImag     : TNvectorPtr);

{----------------------------------------------------}
{- Input: NumPoints, XReal, XImag, HReal, HImag     -}
{- Output: XReal, XImag                             -}
{-                                                  -}
{- This procedure performs the following operation: -}
{-                                                  -}
{-     Dummy[f] := X[f] * H[-f]                     -}
{-                                                  -}
{- where -f is represented by NumPoints - f         -}
{- (circular correlation).                          -}
{-                                                  -}
{- The product is returned in X.                    -}
{----------------------------------------------------}

var
  Term : integer;
  NumPointsMinusTerm : integer;
  DummyReal : TNvectorPtr;
  DummyImag : TNvectorPtr;

begin
  New(DummyReal);
  New(DummyImag);
  DummyReal^[0] := XReal^[0] * HReal^[0] - XImag^[0] * HImag^[0];
  DummyImag^[0] := XReal^[0] * HImag^[0] + XImag^[0] * HReal^[0];
  for Term := 1 to NumPoints - 1 do
  begin
    NumPointsMinusTerm := NumPoints - Term;
    DummyReal^[Term] := XReal^[Term] * HReal^[NumPointsMinusTerm] -
                        XImag^[Term] * HImag^[NumPointsMinusTerm];
    DummyImag^[Term] := XImag^[Term] * HReal^[NumPointsMinusTerm] +
                        XReal^[Term] * HImag^[NumPointsMinusTerm];
  end;
  XReal^ := DummyReal^;
  XImag^ := DummyImag^;
  Dispose(DummyReal);
  Dispose(DummyImag);
end; { procedure Multiply }

begin { procedure ComplexCorrelation }
  TestInput(NumPoints, NumberOfBits, Error);
  if Error = 0 then
  begin
    New(SinTable);
    New(CosTable);
    MakeSinCosTable(NumPoints, SinTable, CosTable);

    { Take the transform of XReal, XImag  }
    Inverse := false;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);

    if not Auto then
      { Take the transform of HReal, HImag  }
      begin
        FFT(NumberOfBits, NumPoints, Inverse, HReal, HImag, SinTable, CosTable);
      end
    else
      { Autocorrelation; set H equal to X  }
      begin
        HReal := XReal;
        HImag := XImag;
      end;

    { Multiply the two transforms together  }
    Multiply(NumPoints, XReal, XImag, HReal, HImag);

    { Take the inverse transform of the product  }
    Inverse := true;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);
    Dispose(SinTable);
    Dispose(CosTable);
  end;
end; { procedure ComplexCorrelation }
