/* - Sample functions to program the DSP32C  Software Development card.
 *
 * Last change: 8/01/89   V 0.2
*/
#include <stdio.h>
#include <dos.h>
#include "coff.h"
#include "dsp32cds.h"


static unsigned int dspbase = BASE;	/* address of the development card */
static unsigned int dspmode = MMODE6;	/* memory mode to run in */


int load_dsp( char *dlfile)
{
	FILHDR *fhdr;
	SCNHDR  *shdrs, *cscn;
	int	i;
	FILE *fp;
	long bytes_to_go;
	unsigned char byte0, byte1;
	unsigned short int data;

	if ((fhdr = (FILHDR *) calloc(1,FILHSZ)) == NULL) {
		fprintf(stderr,"Unable to obtain memory for file header\n");
		return 1;
	}

	if ((fp = fopen(dlfile,"rb")) == NULL) {
		fprintf(stderr,"Unable to open file: %s\n",dlfile);
		return 1;
	}

	/* Read the file header */
	if (fread(fhdr, FILHSZ, 1, fp) != 1) {
		fprintf(stderr,"Unable to read file header\n");
		fclose(fp);
		return 1;
	}

	/* Check the magic number to be sure this is a DSP32/C file */
	if ( fhdr->f_magic != DSP32MAGIC && fhdr->f_magic != DSP32CMAGIC ) {
		fprintf(stderr,"This is not a DSP32/32C object file.\n");
		fclose(fp);
		return 1;
	}

	/* Make sure this file has been linked and all externals resolved */
	if ((fhdr->f_flags & F_EXEC) == 0) {
		fprintf(stderr,"This DSP32/32C file is not executable.\n");
		fclose(fp);
		return 1;
	}

#ifdef DEBUG
	printf("File contains %d sections\n",fhdr->f_nscns);
#endif

	if ((shdrs = (SCNHDR *) calloc(fhdr->f_nscns, SCNHSZ)) == NULL) {
		fprintf(stderr,"Unable to obtain memory for section headers\n");
		return 1;
	}

	if (fhdr->f_opthdr != 0)
		fseek(fp, fhdr->f_opthdr, 1); 	/* skip the optional header if any */


	for ( i = 0, cscn = shdrs; i < fhdr->f_nscns; i++, cscn++ ) {
		if (fread(cscn, SCNHSZ, 1, fp) != 1) {
			fprintf(stderr,"Unable to read section header\n");
			fclose(fp);
			return 1;
		}
#ifdef DEBUG
		printf("sections: %-10s at 0x%.6lx length 0x%.6lx\n",
				cscn->s_name, cscn->s_paddr, cscn->s_size);
#endif
	}
	
	fprintf(stderr,"\nDownloading ...\n");
	for ( i = 0, cscn = shdrs; i < fhdr->f_nscns; i++, cscn++ ) {

		/* make sure the section contains valid data */
		if ( *cscn->s_name == '\0' || cscn->s_size <= 0
			|| cscn->s_scnptr == 0) 
			continue;

		if (fseek(fp, cscn->s_scnptr, 0) < 0 ) {
			fprintf(stderr,"Error unable to seek to section %s\n",
				cscn->s_name);
			fclose(fp);
			return 1;
		}
		fprintf(stderr,"\tsection: %-10s to 0x%.6lx size 0x%.6lx\n",
				cscn->s_name, cscn->s_paddr, cscn->s_size);
		/* Set DMA address and autoincrement mode */
		write_pcr( FLG | AUTO | DMA | ENI | REGMAP);
		write_par(cscn->s_paddr);

		/* Whenever DMA is turned on the DSP32C load the PDR with
		* data from memory. To avoid a conflict with between input
		* and output DMA so wait unit the read DMA is complete.
		* This happens so fast that this isnt needed but lets be
		* safe.
		*/
		while((read_pcr() & PDFS) == 0);

		bytes_to_go = cscn->s_size;
		while ( bytes_to_go > 0 )  {
			byte0 = fgetc(fp);
			byte1 = fgetc(fp);
			data = byte1 << 8 | byte0;
			write_pdr16(data);
			/* wait for dma to complete */
			while((read_pcr() & PDFS) != 0);
			bytes_to_go -= 2;
		}
	}
	return 0;
		

}

void
reset_dsp()
{
}

/* Halt the chip  with DMA enabled */
void
halt_dsp()
{
	write_pcr(REGMAP | ENI | AUTO | DMA | FLG);
}

/* Set the DSP running assume its halted */
void
run_dsp()
{
	/* First write the PCRL causing a 0 to 1 transition of the RESET bit.
	*  This clears all the other PCR bits.
	*  Next write PCRL to set the bits that were cleared so we can then
	*  access PCRH.
	*/
	outportb(PCRL, ENI | REGMAP| RESET);
	/* Set previously cleared bits */
	write_pcr( FLG | ENI | REGMAP | RESET );
}

int
init_card()
{
	write_pcr(FLG | ENI | REGMAP );	/* halt the DSP */

	/* Toggle INTREQ2 */
	outportb(INITWR, dspmode | CONTZN | IRQENABLE | IRQ | BPCNTL );

	/* this is to init the optional DRAM card
	to assert SRDYN run the chip and then stop it */
	write_pcr(FLG | ENI | REGMAP | RESET );	/* run the chip for a few bit */
	write_pcr(FLG | ENI | REGMAP );	/* and then stop it */

	outportb(INITWR, dspmode | CONTZN | IRQENABLE | BPCNTL );
	outportb(INITWR, dspmode | CONTZN | IRQENABLE | IRQ | BPCNTL );
	/* Deassert ZN and enable interrupt logic to SIO Header */
	outportb(INITWR, dspmode | CONTZN | IRQ );
	
	if ((read_pcr() & 0x9) != 0 ) {
		printf("Check Hardward Configuration\n");
		printf("No DSP32C Software Development Card found at 0x%x\n",dspbase);
		return 1;
	}


	/* set CMMD2 if MMD2 is 1 so PAL will decode externel memory */
	if ( dspmode > 3 )
		outportb(CNTLWR2, CMMD2);
	else
		outportb(CNTLWR2, 0);

	if (test4_dsp()) 
		return 1;

	return 0;
}

/* Write 16 bit value to the PCR register */
void
write_pcr( unsigned short int data)
{
#ifdef DEBUG
	printf("write_pcr: 0x%.4x\n",data);
#endif
	outportb( PCRL, data);
	outportb( PCRH, data>>8);
}

/* Read 16 bit value from the PCR register */
unsigned short int
read_pcr() {
	unsigned short low, high, data;

	low = inportb(PCRL);
	high = inportb(PCRH);
	data = (low & 0xff) | ( high << 8 );
#ifdef DEBUG
	printf("read_pcr: 0x%.4x\n",data);
#endif
	return data;
}

/* Read 16 bit data the PDR register */
unsigned short
read_pdr16() 
{
	unsigned short low, high, data;
	low = inportb(PDRL);
	high = inportb(PDRH);
	data = (high << 8) | (low & 0xff);
#ifdef DEBUG
	printf("read_pd16: 0x%.4x\n",data);
#endif
	return data;
}

/* Write 16 bit data the PDR register */
void
write_pdr16( unsigned short data)
{
#ifdef DEBUG
	printf("write_pdr16: 0x%.4x\n",data);
#endif
	outportb( PDRL, data);
	outportb( PDRH, data>>8);
}

/* Read 32 bit data the PDR register */
unsigned long
read_pdr32() 
{
	unsigned char b1, b2, b3, b4;
	unsigned long data;
	b3 = inportb(PDR2L);
	b4 = inportb(PDR2H);
	b1 = inportb(PDRL);
	b2 = inportb(PDRH);
	data  = (long)b4<<24 | (long)b3<<16 | (long)b2<<8 | (long)b1;
#ifdef DEBUG
	printf("read_pdr32: 0x%.8lx\n",data);
#endif
	return data;
}

/* Write 32 bit data the PDR register */
void
write_pdr32( unsigned long data)
{
#ifdef DEBUG
	printf("write_pdr32: 0x%.8lx\n",data);
#endif
	outportb( PDR2L, data>>16);
	outportb( PDR2H, data>>24);
	outportb( PDRL, data);
	outportb( PDRH, data>>8);
}

/* Write 24 bit PAR register */
void
write_par( unsigned long address)
{
	unsigned char dsp_pare, dsp_parl, dsp_parh;
#ifdef DEBUG
	printf("write_par: 0x%.6lx\n",address);
#endif
	dsp_parl = address;
	dsp_parh = address>>8;
	dsp_pare = address>>16;
	outportb( PARE, dsp_pare);
	outportb( PARL, dsp_parl);
	outportb( PARH, dsp_parh);
}


void
dl_array( unsigned long address, unsigned int num, int *array)
{
/* Download an array of integers to the DSP32C starting at 'address'.
 * 'num' holds the number of integers.  'array' is a pointer to the
 * array of integers.
 */
	/* write starting address to the parallel address register */
	write_par(address);
	/* Wait for initial DMA read to complete before writing data */
	while ((read_pcr() & PDFS) == 0);

	while(num-- != 0)  {
		write_pdr16(*array++);
		/* wait for write to complete*/
		while ((read_pcr() & PDFS) != 0 );
	}
}

void
ul_array(unsigned long address, unsigned int num, int *array) {
/*
 * Upload an array from the DSP32C memory to the PC.
 */

	/* write starting address to the parallel address register */
	write_par( address );

	while( num-- != 0 ) {
		while ((read_pcr() & PDFS) == 0);
		*array++ = read_pdr16();
	}
}


/* Enable codec operation */
void codec_on()
{
	outportb(CNTLWR, 0);
	if ( dspmode > 3 )
		outportb(CNTLWR2, CODBENH | CMMD2);
	else
		outportb(CNTLWR2, CODBENH);
}

/* Set the base address of the DSP32C-DS */
void set_base(unsigned addr)
{
	dspbase = addr;
}

/* Set the memory mode of the DSP32C-DS */
void set_memmode(unsigned int mode)
{
	dspmode = (mode ^ 2) << 3;
}

/* test for functioning dsp32c card */
test4_dsp()
{
	/* test for a DSP32C by DMA in memory and checking
	* DMA out of correct values.
	*/
	static int in[5] = { 0x1234, 0, 0xffff, 0xaaaa, 0x5555 };
	int out[5];
	int i;
	
	write_pcr( FLG | AUTO | DMA | ENI | REGMAP);
	dl_array(0L, 5, in);
	ul_array(0L, 5, out);
	write_pcr( FLG | ENI | REGMAP);
	for ( i = 0 ; i < 5 ; i++ )
		if ( in[i] != out[i] ) {
			printf("Check Hardward Configuration\n");
			printf("No DSP32C Software Development Card found at 0x%x\n",dspbase);
			return 1;
		}
	return 0;
}

