procedure RealConvolution(NumPoints : integer;
                      var XReal     : TNvectorPtr;
                      var XImag     : TNvectorPtr;
                      var HReal     : TNvectorPtr;
                      var Error     : byte);

{-------------------------------------------------------------------}
{-                                                                 -}
{-     Turbo Pascal Numerical Methods Toolbox                      -}
{-     (C) Copyright 1986 Borland International.                   -}
{-                                                                 -}
{-   Input: NumPoints, XReal, XImag, HReal                         -}
{-   Output: XReal, XImag, Error                                   -}
{-                                                                 -}
{-   Purpose: This procedure performs a convolution of the         -}
{-            real data XReal and HReal.  The result is returned   -}
{-            in the complex vector XReal, XImag.                  -}
{-                                                                 -}
{- User Defined Types:                                             -}
{-         TNvector = array[0..TNArraySize] of real                -}
{-      TNvectorPtr = ^TNvector                                    -}
{-                                                                 -}
{- Global Variables:  NumPoints : integer     Number of data       -}
{-                                            points in X          -}
{-                    XReal     : TNvectorPtr Data points          -}
{-                    HReal     : TNvectorPtr Data points          -}
{-                    Error     : byte        Indicates an error   -}
{-                                                                 -}
{-             Errors:  0: No Errors                               -}
{-                      1: NumPoints < 2                           -}
{-                      2: NumPoints not a power of two            -}
{-                                                                 -}
{-   Version Date:  26 January 1987                                -}
{-                                                                 -}
{-------------------------------------------------------------------}

var
  SinTable, CosTable : TNvectorPtr;    { Tables of sine and cos values  }
  Inverse : boolean;                   { Indicates inverse transform  }
  NumberOfBits : byte;                 { Number of bits required to  }
                                       { represent NumPoints    }
  HImag : TNvectorPtr;                 { Imaginary part of the   }
                                       { FFT transform of HReal  }

procedure Multiply(NumPoints : integer;
               var XReal     : TNvectorPtr;
               var XImag     : TNvectorPtr;
               var HReal     : TNvectorPtr;
               var HImag     : TNvectorPtr);

{----------------------------------------------}
{- Input: NumPoints, XReal, XImag, HReal,     -}
{-        HImag                               -}
{- Output: XReal, XImag                       -}
{-                                            -}
{- This procedure multiplies each element in  -}
{- X by the corresponding element in H.  The  -}
{- product is returned in X.                  -}
{----------------------------------------------}

var
  Term : integer;
  Dummy : real;

begin
  for Term := 0 to NumPoints - 1 do
  begin
    Dummy := XReal^[Term] * HReal^[Term] - XImag^[Term] * HImag^[Term];
    XImag^[Term] := XReal^[Term] * HImag^[Term] +
                    XImag^[Term] * HReal^[Term];
    XReal^[Term] := Dummy;
  end;
end; { procedure Multiply }

procedure SeparateTransforms(NumPoints : integer;                              var XReal : TNvectorPtr;
                         var XImag     : TNvectorPtr;
                         var HReal     : TNvectorPtr;
                         var HImag     : TNvectorPtr);

{-------------------------------------------------------}
{- Input: NumPoints, XReal, XImag                      -}
{- Output: HReal, HImag                                -}
{-                                                     -}
{- The transforms of the real data XReal and HReal are -}
{- stored in the vector XReal, XImag.  By using the    -}
{- symmetries of the transforms of real data, this     -}
{- procedure extracts the complex transforms.  The     -}
{- transform of XReal is returned in XReal, XImag. The -}
{- transform of HReal is returned in HReal, HImag.     -}
{-------------------------------------------------------}

var
  Term : integer;
  EndTerm : integer;
  DummyReal : real;
  DummyImag : real;
  NumPointsSHR1 : integer;
  NumPointsMinusTerm : integer;

begin
  HReal^[0] := XImag^[0];
  HImag^[0] := 0;
  XImag^[0] := 0;
  NumPointsSHR1 := NumPoints SHR 1;
  HReal^[NumPointsSHR1] := XImag^[NumPointsSHR1];
  HImag^[NumPointsSHR1] := 0;
  XImag^[NumPointsSHR1] := 0;
  EndTerm := NumPointsSHR1 - 1;
  for Term := 1 to EndTerm do
  begin
    NumPointsMinusTerm := NumPoints - Term;
    HReal^[Term] := 0.5 * (XImag^[Term] + XImag^[NumPointsMinusTerm]);
    HImag^[Term] := 0.5 * (XReal^[NumPointsMinusTerm] - XReal^[Term]);
    DummyReal := 0.5 * (XReal^[Term] + XReal^[NumPointsMinusTerm]);
    DummyImag := 0.5 * (XImag^[Term] - XImag^[NumPointsMinusTerm]);
    XReal^[Term] := DummyReal;
    XImag^[Term] := DummyImag;
  end;

  for Term := 1 to EndTerm do
  begin  { Make use of symmetries to calculate  }
         { the rest of the transform            }
    NumPointsMinusTerm := NumPoints - Term;
    XReal^[NumPointsMinusTerm] :=  XReal^[Term];
    XImag^[NumPointsMinusTerm] := -XImag^[Term];
    HReal^[NumPointsMinusTerm] :=  HReal^[Term];
    HImag^[NumPointsMinusTerm] := -HImag^[Term];
  end;
end; { procedure SeparateTransforms }

begin { procedure RealConvolution }
  TestInput(NumPoints, NumberOfBits, Error);
  if Error = 0 then
  begin
    New(HImag);
    New(SinTable);
    New(CosTable);

    { Combine the two real transforms  }
    { into one complex transform       }
    XImag^ := HReal^;
    MakeSinCosTable(NumPoints, SinTable, CosTable);

    { Take the transform of XReal, XImag  }
    Inverse := false;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);

    { Separate out the X transform and H transform  }
    SeparateTransforms(NumPoints, XReal, XImag, HReal, HImag);

    { Multiply the two transforms  }
    Multiply(NumPoints, XReal, XImag, HReal, HImag);

    { Take the inverse transform of the product  }
    Inverse := true;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);
    Dispose(SinTable);
    Dispose(CosTable);
    Dispose(HImag);
  end;
end; { procedure RealConvolution }
