procedure RealCorrelation(NumPoints : integer;
                      var Auto      : boolean;
                      var XReal     : TNvectorPtr;
                      var XImag     : TNvectorPtr;
                      var HReal     : TNvectorPtr;
                      var Error     : byte);

{-------------------------------------------------------------------}
{-                                                                 -}
{-     Turbo Pascal Numerical Methods Toolbox                      -}
{-     (C) Copyright 1986 Borland International.                   -}
{-                                                                 -}
{-   Input: NumPoints, Auto, XReal, XImag, HReal                   -}
{-   Output: XReal, XImag, Error                                   -}
{-                                                                 -}
{-   Purpose: This procedure performs a correlation (auto or       -}
{-            cross) of the real data XReal and HReal. The         -}
{-            correlation is returned in the complex vector        -}
{-            XReal, XImag.                                        -}
{-                                                                 -}
{- User Defined Types:                                             -}
{-         TNvector = array[0..TNArraySize] OF real                -}
{-      TNvectorPtr = ^TNvector                                    -}
{-                                                                 -}
{- Global Variables:  NumPoints : integer     Number of data       -}
{-                                            points in X          -}
{-                    Auto      : boolean     True => auto-        -}
{-                                                    correlation  -}
{-                                            False=> cross-       -}
{-                                                    correlation  -}
{-                    XReal     : TNvectorPtr First sample         -}
{-                    HReal     : TNvectorPtr Second sample        -}
{-                    Error     : byte        Indicates an error   -}
{-                                                                 -}
{-             Errors:  0: No Errors                               -}
{-                      1: NumPoints < 2                           -}
{-                      2: NumPoints not a power of two            -}
{-                                                                 -}
{-   Version Date: 26 January 1987                                 -}
{-                                                                 -}
{-------------------------------------------------------------------}

var
  SinTable, CosTable : TNvectorPtr;  { Tables of sine and cosine values  }
  Inverse : boolean;                 { Indicates inverse transform  }
  NumberOfBits : byte;               { Number of bits necessary to     }
                                     { represent the number of points  }
  HImag : TNvectorPtr;               { Imaginary part of the   }
                                     { FFT transform of HReal  }

procedure SeparateTransforms(NumPoints : integer;
                         var XReal     : TNvectorPtr;
                         var XImag     : TNvectorPtr;
                         var HReal     : TNvectorPtr;
                         var HImag     : TNvectorPtr);

{-------------------------------------------------------}
{- Input: NumPoints, XReal, XImag                      -}
{- Output: HReal, HImag                                -}
{-                                                     -}
{- The transforms of the real data XReal and HReal are -}
{- stored in the vector XReal, XImag.  By using the    -}
{- symmetries of the transforms of real data, this     -}
{- procedure extracts the complex transforms.  The     -}
{- transform of XReal is returned in XReal, XImag. The -}
{- transform of HReal is returned in HReal, HImag.     -}
{-------------------------------------------------------}

var
  Term : integer;
  EndTerm : integer;
  DummyReal : real;
  DummyImag : real;
  NumPointsSHR1 : integer;
  NumPointsMinusTerm : integer;

begin
  HReal^[0] := XImag^[0];
  HImag^[0] := 0;
  XImag^[0] := 0;
  NumPointsSHR1 := NumPoints SHR 1;
  HReal^[NumPointsSHR1] := XImag^[NumPointsSHR1];
  HImag^[NumPointsSHR1] := 0;
  XImag^[NumPointsSHR1] := 0;
  EndTerm := NumPointsSHR1 - 1;
  for Term := 1 to EndTerm do
  begin
    NumPointsMinusTerm := NumPoints - Term;
    HReal^[Term] := 0.5 * (XImag^[Term] + XImag^[NumPointsMinusTerm]);
    HImag^[Term] := 0.5 * (XReal^[NumPointsMinusTerm] - XReal^[Term]);
    DummyReal := 0.5 * (XReal^[Term] + XReal^[NumPointsMinusTerm]);
    DummyImag := 0.5 * (XImag^[Term] - XImag^[NumPointsMinusTerm]);
    XReal^[Term] := DummyReal;
    XImag^[Term] := DummyImag;
  end;

  for Term := 1 to EndTerm do
  begin  { Make use of symmetries to calculate  }
         { the rest of the transform            }
    NumPointsMinusTerm := NumPoints - Term;
    XReal^[NumPointsMinusTerm] :=  XReal^[Term];
    XImag^[NumPointsMinusTerm] := -XImag^[Term];
    HReal^[NumPointsMinusTerm] :=  HReal^[Term];
    HImag^[NumPointsMinusTerm] := -HImag^[Term];
  end;
end; { procedure SeparateTransforms }

procedure Multiply(NumPoints : integer;
               var XReal     : TNvectorPtr;
               var XImag     : TNvectorPtr;
               var HReal     : TNvectorPtr;
               var HImag     : TNvectorPtr);

{----------------------------------------------------}
{- Input: NumPoints, XReal, XImag, HReal, HImag     -}
{- Output: XReal, XImag                             -}
{-                                                  -}
{- This procedure performs the following operation: -}
{-                                                  -}
{-     Dummy[f] := X[f] * H[-f]                     -}
{-                                                  -}
{- where -f is represented by NumPoints - f         -}
{- (circular correlation).  Because x and h were    -}
{- real functions, this operation is identical      -}
{- to:                                              -}
{-                           *                      -}
{-    Dummy[f] := X[f] - H[f]                       -}
{-                                                  -}
{- The product is returned in X.                    -}
{----------------------------------------------------}

var
  Term : integer;
  NumPointsMinusTerm : integer;
  DummyReal : TNvectorPtr;
  DummyImag : TNvectorPtr;

begin
  New(DummyReal);
  New(DummyImag);
  DummyReal^[0] := XReal^[0] * HReal^[0] - XImag^[0] * HImag^[0];
  DummyImag^[0] := XReal^[0] * HImag^[0] + XImag^[0] * HReal^[0];
  for Term := 1 to NumPoints - 1 do
  begin
    NumPointsMinusTerm := NumPoints - Term;
    DummyReal^[Term] := XReal^[Term] * HReal^[NumPointsMinusTerm] -
                        XImag^[Term] * HImag^[NumPointsMinusTerm];
    DummyImag^[Term] := XImag^[Term] * HReal^[NumPointsMinusTerm] +
                        XReal^[Term] * HImag^[NumPointsMinusTerm];
  end;
  XReal^ := DummyReal^;
  XImag^ := DummyImag^;
  Dispose(DummyReal);
  Dispose(DummyImag);
end; { procedure Multiply }

begin { procedure RealCorrelation }
  TestInput(NumPoints, NumberOfBits, Error);
  if Error = 0 then
  begin
    if Auto then
      FillChar(XImag^, SizeOf(XImag^), 0)
    else    { Combine the two real transforms  }
            { into one complex transform       }
      XImag^ := HReal^;
    New(HImag);
    New(SinTable);
    New(CosTable);
    MakeSinCosTable(NumPoints, SinTable, CosTable);

    { Take the transform of XReal, XImag  }
    Inverse := false;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);

    if Auto then
      begin
        HReal^ := XReal^;
        HImag^ := XImag^;
      end
    else
      SeparateTransforms(NumPoints, XReal, XImag, HReal, HImag);

    { Multiply the two transforms together  }
    Multiply(NumPoints, XReal, XImag, HReal, HImag);

    { Take the inverse transform of the product  }
    Inverse := true;
    FFT(NumberOfBits, NumPoints, Inverse, XReal, XImag, SinTable, CosTable);
    Dispose(SinTable);
    Dispose(CosTable);
    Dispose(HImag);
  end;
end; { procedure RealCorrelation }
