{************************************************************************** 
 
   ADSP-2101 Family Software UART                             UART.DSP 
     
    This uses FLAG_IN, FLAG_OUT and the TIMER of ADSP-2101 to interface to  
    an RS-232 asynchronous serial device such as a VT100 terminal.  ex: 
     
	ADSP-2101 FLAG_OUT ----------> AD233 ----------> RS-232 RX 
	 
	ADSP-2101  FLAG_IN <---------- AD233 <---------- RS-232 TX 
		       (TIMER maintains baudrate) 
	 
    Parameters bits/word, baudrate, stopbits & parity are user-programmable. 
     
    An RS-232 line driver chip (such as the AD233) can be used to electrically  
    interface +5 VDC to the RS-232 line voltage levels.  
 
    The operation of the transmitter setup routine is completely  
    independent on the the receiver setup routine operation.  Although  
    both tx and rx use the same timer as a master clock source, the 
    xmitted bits need not be in sync with the received bits. The default 
    state of the reciever is OFF, so the "turn_rx_on" subroutine must be 
    used to enable RX. 
 
Calling Argument: 
	     for autobaud load the baud constant 
	dm(baud_period)=(Proc_frequency/(3*Baudrate))-1 
 
Useful Subroutines: 
	init_uart       Must be called after system reset. 
	get_char_ax1    Waits for RX input and returns with it in ax1. 
	out_char_ax1    Waits for last TX output and transmits data from ax1. 
	turn_rx_on      Must be called to enable the receipt of RX data. 
	turn_rx_off     Can be used to ignore input RX data. 
 
Useful Flag: 
	DM(flag_rx_ready)   If this DM location is all ones it indicates that 
			    the UART is ready to rx new word. If it is zero 
			    then data is being received. Can be used for xon 
			    xoff flow control. 
 
    Author:   Fares Eidi, 21-May-90, Analog Devices Inc. 
    modified: Christoph D. Cavigioli, 17-Dec-90 
    modified: Steven Cox, 20-Dec-91, extensive rewrite 
    modified: Steven Cox, 31-Dec-91, Reset flag_rx_stop_yet to 1, Comments. 
    modified: Steven Cox, 11-Feb-92, Added support for autobaud. 
    modified: Philip Holdgate, 02-Apr-92, modified for Apps Note, cleaned up. 
 
****************************************************************************} 
 
 
.module/boot=1              UART; 
 
 
{______The Constants below must be changed to modify uart parameters _______} 
 
.const  tx_num_of_bits = 10;    { start bits + tx data bits + stop bits } 
.const  rx_num_of_bits = 8;     { rx data bits (start&stop bits not counted)} 
.const  RX_BIT_ADD = 0x0100;    { = 1<<rx_num_of_bits } 
.const  TX_BIT_ADD = 0xfe00;    { = 0xffff<<(tx data bits+1) } 
        
       {....These constants can be used if autobaud is not needed....} 
 
.const  PERIOD = 74;    {13&57600}{ PERIOD = (Proc_frequency/(3*Baudrate))-1} 
{.const  PERIOD = 112;} {13&38400}{ PERIOD = (Proc_frequency/(3*Baudrate))-1} 
{.const  PERIOD = 225;} {13&19200}{ PERIOD = (Proc_frequency/(3*Baudrate))-1} 
{.const  PERIOD = 450;} {13& 9600}{ PERIOD = (Proc_frequency/(3*Baudrate))-1} 
 
{________________Definitions of memory mapped control registers____________} 
 
.const  TSCALE=                 0x3ffb; 
.const  TCOUNT=                 0x3ffc; 
.const  TPERIOD=                0x3ffd; 
.const  System_Control_Reg=     0x3fff; 
 
{___________________________________________________________________________} 
 
.entry init_uart;         { UART initialize baudrate etc. } 
.entry out_char_ax1;      { UART output a character } 
.entry get_char_ax1;      { UART wait & get input character } 
.entry turn_rx_on;        { UART enable the rx section } 
.entry turn_rx_off;       { UART disable the rx section } 
.entry process_a_bit;     { UART timer interrupt routine for RX and TX } 
 
.global flag_rx_ready; 
.global baud_period; 
     
.var    flag_tx_ready;      { flag indicating UART is ready for new tx word } 
.var    flag_rx_ready;      { flag indicating UART is ready to rx new word } 
.var    flag_rx_stop_yet;   { flag tells that a rx stop bit is not pending } 
.var    flag_rx_no_word;    { indicates a word is not in the user_rx_buffer } 
.var    flag_rx_off;        { indicates a that the receiver is turned off } 
.var    timer_tx_ctr;       { divide by 3 ctr, timer is running @ 3x baudrate } 
.var    timer_rx_ctr;       { divide by 3 ctr, timer is running @ 3x baudrate } 
.var    user_tx_buffer;     { UART tx reg loaded by user before UART xmit } 
.var    user_rx_buffer;     { UART rx reg read by user after word is rcvd } 
.var    internal_tx_buffer; { formatted for serial word, adds start&stop bits } 
			    { 'user_tx_buffer' is copied here before xmission } 
.var    internal_rx_buffer; 
.var    bits_left_in_tx;  { number of bits left in tx buffer (not yet clkd out) } 
.var    bits_left_in_rx;  { number of bits left to be rcvd (not yet clkd in) } 
.var    baud_period;        { loaded by autobaud routine } 
 
 
{_________________________Initializing subroutine_________________________} 
 
init_uart: 
    ax0=0;       
    dm(TSCALE)=ax0;             { decrement TCOUNT every instruction cycle } 
 
    ax0=dm(baud_period);        { from autobaud or use constant: ax0=PERIOD; } 
				{ ...and comment in the appropriate constant } 
    dm(TCOUNT)=ax0; 
    dm(TPERIOD)=ax0;            { interrupts generated at 3x baudrate } 
    ax0=0; 
    dm(System_Control_Reg)=ax0; { no bmwait, pmwait states, SPORT1 = FI/FO } 
     
    ax0=1;     
    dm(flag_tx_ready)=ax0;      { set the flags showing that UART is not busy } 
    dm(flag_rx_ready)=ax0; 
    dm(flag_rx_stop_yet)=ax0; 
    dm(flag_rx_no_word)=ax0; 
    dm(flag_rx_off)=ax0;        { rx section off } 
 
    set flag_out;               { UART tx output is initialized to high } 
    ifc=0x003f;                 { clear all pending interrupts } 
    nop;                        { wait for ifc latency } 
    imask=b#000001;             { enable TIMER interrupt handling } 
    ena timer;                  { start timer now } 
    rts; 
 
{___________________Process_a_bit (TIMER interrupt routine)_________________ 
  
    This routine is the heart of the UART.  It is called every timer  
    interrupt (i.e. 3x baudrate).  This routine will xmit one bit at a  
    time by setting/clearing the FLAG_OUT pin of the ADSP-2101. This routine 
    will then test if the uart is already receiving. If not it will test  
    flagin (rx) for a start bit and place the uart in receive mode if true. 
    If already in receive mode it will shift in one bit at a time by reading  
    the FLAG_IN pin.  Since the internal timer is running at 3x baudrate,  
    bits need only be transmitted/received once every 3 timer interrupts. 
_____________________________________________________________________________} 
process_a_bit: 
    ena sec_reg;                    { Switch to the background dreg set } 
    ax0=dm(flag_tx_ready);     { if not in "transmit", go right to "receive" } 
    ar=pass ax0;                     
    if ne jump receiver;             
	       {_________Transmitter Section________} 
    ay0=dm(timer_tx_ctr);           { test timer ctr to see if a bit } 
    ar=ay0-1;                       { is to be sent this time around } 
    dm(timer_tx_ctr)=ar;            { if no bit is to be sent } 
    if ne jump receiver;            { then decrement ctr and return } 
 
    sr1=dm(internal_tx_buffer);     { shift out LSB of internal_tx_buffer } 
    sr=lshift sr1 by -1 (hi);       { into SR1.  Test the sign of this bit } 
    dm(internal_tx_buffer)=sr1;     { set or reset FLAG_OUT accordingly } 
    ar=pass sr0;                    { this effectively clocks out the } 
    if ge reset flag_out;           { word being xmitted one bit at a time }  
    if lt set flag_out;             { LSB out first at FLAG_OUT. } 
     
    ay0=3;                          { reset timer ctr to 3, i.e. next bit } 
    dm(timer_tx_ctr)=ay0;           { will be sent after 3 timer interrupts } 
 
    ay0=dm(bits_left_in_tx);        { number of bits left to be xmitted } 
    ar=ay0-1;                       { is now decremented by one, } 
    dm(bits_left_in_tx)=ar;         { indicating that one is now xmitted } 
    if gt jump receiver;            { if no more bits left, then ready } 
 
    ax0=1;                          { flag is set to true indicating } 
    dm(flag_tx_ready)=ax0;          { a new word can now be xmitted } 
 
	       {_________Receiver Section________} 
receiver: 
    ax0=dm(flag_rx_off);            { Test if receiver is turned on } 
    ar=pass ax0; 
    if ne rti; 
 
    ax0=dm(flag_rx_stop_yet);       { Test if finished with stop bit of } 
    ar=pass ax0;                    { last word or not. if finished then } 
    if ne jump rx_test_busy;        { continue with check for receive. } 
 
    ay0=dm(timer_rx_ctr);           { decrement timer ctr and test to see } 
    ar=ay0-1;                       { if stop bit period has been reached } 
    dm(timer_rx_ctr)=ar;            { if not return and wait } 
    if ne rti; 
 
    ax0=1;                          { if stop bit is reached then reset } 
    dm(flag_rx_stop_yet)=ax0;       { to wait for next word } 
    dm(flag_rx_ready)=ax0; 
 
    ax0=dm(internal_rx_buffer);     { copy internal rx buffer } 
    dm(user_rx_buffer)=ax0;         { to the user_rx_buffer } 
 
    ax0=0;                          { indicated that a word is ready in } 
    dm(flag_rx_no_word)=ax0;        { the user_rx_buffer } 
    rti; 
 
rx_test_busy: 
    ax0=dm(flag_rx_ready);          { test rx flag, if rcvr is not busy } 
    ar=pass ax0;                    { receiving bits then test for start.If it} 
    if eq jump rx_busy;             { is busy, then clk in one bit at a time } 
 
    if flag_in jump rx_exit;        { Test for start bit and return if none } 
 
    ax0=0; 
    dm(flag_rx_ready)=ax0;          { otherwise, indicate rcvr is now busy } 
    dm(internal_rx_buffer)=ax0;     { clear out rcv register} 
     
    ax0=4;                          { timer runs @ 3x baud rate, so rcvr } 
    dm(timer_rx_ctr)=ax0;           { will only rcv on every 3rd interrupt } 
				    { initially this ctr is set to 4.  This } 
				    { will skip the start bit and will } 
				    { allow us to check FLAG_IN at the center } 
				    { of the received data bit } 
    ax0=rx_num_of_bits; 
    dm(bits_left_in_rx)=ax0; 
rx_exit: 
    rti; 
 
rx_busy: 
    ay0=dm(timer_rx_ctr);           { decrement timer ctr and test to see } 
    ar=ay0-1;                       { if bit is to be rcvd this time around } 
    dm(timer_rx_ctr)=ar;            { if not return, else receive a bit } 
    if ne rti; 
 
rcv:            { Shift in rx bit } 
    ax0=3;                          { reset the timer ctr to 3 indicating } 
    dm(timer_rx_ctr)=ax0;           { next bit is 3 timer interrupts later } 
     
    ay0=RX_BIT_ADD;             
    ar=dm(internal_rx_buffer); 
    if not flag_in jump pad_zero;   { Test RX input bit and } 
    ar=ar+ay0;                      { add in a 1 if hi } 
 
pad_zero: 
    sr=lshift ar by -1 (lo);        { Shift down to ready for next bit } 
    dm(internal_rx_buffer)=sr0; 
 
    ay0=dm(bits_left_in_rx);        { if there are more bits left to be rcvd } 
    ar=ay0-1;                       { then keep UART in rcv mode } 
    dm(bits_left_in_rx)=ar;         { and return } 
    if gt rti;                      { if there are no more bits then.. } 
 
		       { That was the last bit } 
    ax0=3;                          { set timer to wait for middle of the } 
    dm(timer_rx_ctr)=ax0;           { stop bit } 
    ax0=0;                          { flag indicated that uart is waiting } 
    dm(flag_rx_stop_yet)=ax0;       { for the stop bit to arrive } 
    rti; 
 
{___________________invoke_UART_transmit subroutine____________________ 
 
    This is the first step in the transmit process.  The user has now 
    loaded 'user_tx_buffer' with the ascii code and has also invoked 
    this routine. 
__________________________________________________________________________} 
invoke_UART_transmit: 
    ax0=3;                          { initialize the timer decimator ctr } 
    dm(timer_tx_ctr)=ax0;           { this divide by three ctr is needed } 
				    { since timer runs @ 3x baud rate } 
 
    ax0=tx_num_of_bits;             { this constant is defined by the } 
    dm(bits_left_in_tx)=ax0;        { user and represents total number of } 
				    { bits including stop and parity } 
				    { ctr is initialized here indicating }     
				    { none of the bits have been xmitted } 
    sr1=0; 
    sr0=TX_BIT_ADD;                 { upper bits are hi to end txmit with hi } 
    ar=dm(user_tx_buffer);          { transmit register is copied into  } 
    sr=sr or lshift ar by 1 (lo);   { the internal tx reg & left justified }  
    dm(internal_tx_buffer)=sr0;     { before it gets xmitted } 
 
    ax0=0;                          { indicate that the UART is busy } 
    dm(flag_tx_ready)=ax0;           
    rts; 
     
{______________________get an input character______________________________ 
output:         ax1 
modifies:       ax0 
___________________________________________________________________________} 
get_char_ax1: 
	ax0=dm(flag_rx_no_word); 
	ar=pass ax0; 
	if ne jump get_char_ax1;       { if no rx word input, then wait } 
 
	ax1=dm(user_rx_buffer);         { get received ascii character } 
	ax0=1; 
	dm(flag_rx_no_word)=ax0;        { word was read } 
	rts; 
 
{______________________output a character__________________________________ 
input:          ax1 
modifies:       ax0, sr1, sr0, ar 
___________________________________________________________________________} 
out_char_ax1: 
	ax0=dm(flag_tx_ready); 
	ar=pass ax0; 
	if eq jump out_char_ax1;   { if tx word out still pending, then wait } 
	dm(user_tx_buffer)=ax1; 
	call invoke_UART_transmit;      { send it out } 
	rts; 
 
{______________________enable the RX section________________________________ 
modifies:       ax0 
___________________________________________________________________________} 
turn_rx_on: 
	ax0=0; 
	dm(flag_rx_off)=ax0; 
	rts; 
 
{______________________disable the RX section_______________________________ 
modifies:       ax0 
___________________________________________________________________________} 
turn_rx_off: 
	ax0=1; 
	dm(flag_rx_off)=ax0; 
	rts; 
 
.endmod; 
